@@ -7,6 +7,9 @@
  */
 
 extern uint32_t lcg_seed;
+#ifdef USE_OPENMP
+#pragma omp threadprivate(lcg_seed)
+#endif
 
 static inline uint32_t
 lcg_rand (void)
@@ -27,6 +30,13 @@ lcg_rand_n (int max)
     return lcg_rand () % max;
 }
 
+static inline uint32_t
+lcg_rand_N (int max)
+{
+    uint32_t lo = lcg_rand ();
+    uint32_t hi = lcg_rand () << 15;
+    return (lo | hi) % max;
+}
 
 /* CRC 32 computation
  */
@@ -43,3 +53,15 @@ image_endian_swap (pixman_image_t *img, int bpp);
 /* Generate n_bytes random bytes in malloced memory */
 uint8_t *
 make_random_bytes (int n_bytes);
+
+/* main body of the fuzzer test */
+int
+fuzzer_test_main (const char *test_name,
+		  int         default_number_of_iterations,
+		  uint32_t    expected_checksum,
+		  uint32_t    (*test_function)(int testnum, int verbose),
+		  int         argc,
+		  const char *argv[]);
+
+void
+fail_after (int seconds, const char *msg);
